/*
 * CudaDistanceMatrix.cu
 *
 * Distance kernel implementation
 *
 *      Author: Pablo David Gutiérrez
 *      Author: Miguel Lastra
 *      Author: Jaume Barcardit
 *      Author: José Manuel Benítez
 *      Author: Francisco Herrera
 */

#include "CudaDistanceMatrix.h"

template<class T>
struct SharedMemory
{
    __device__ inline operator       T*()
    {
        extern __shared__ int __smem[];
        return (T*)__smem;
    }

    __device__ inline operator const T*() const
    {
        extern __shared__ int __smem[];
        return (T*)__smem;
    }
};

__global__ void distanceKernel(float* training, float* test, int nVariables, int trainingSize, int testSize, float* matrix, int chunkTrain, int nChunksTrain, int nNumeric, int selections){
	int trainId = threadIdx.x;
	int testId = blockIdx.x;

	float trainVal, testVal;
	float tmpResult;

	/* Setting pointers to their  positions in memory*/
	float* m = &matrix[testId*(TRAININGCHUNK+selections)];

	//Copy the test values to shared memory
	float* shared = SharedMemory<float>();
	for(int i = threadIdx.x; i < nVariables; i+=blockDim.x)
		shared[i] = test[i*TESTCHUNK+testId];

	__syncthreads();

	//Compute NREP distances
	for(int r = 0; r < NREP;r++){
		float result = 0;
		//Euclidean distance
		for(int i = 0; i < nNumeric; i++){
			trainVal = training[i*TRAININGCHUNK+trainId];
			testVal = shared[i];
			tmpResult = (trainVal - testVal);
			result += tmpResult * tmpResult;
		}
		//A different approach for categorical attributes distance can be used
		//Euclidean distance
		for(int i = nNumeric; i < nVariables-1; i++){
			trainVal = training[i*TRAININGCHUNK+trainId];
			testVal = shared[i];
			//result += (trainVal == testVal) ? 0 : 1;
			tmpResult = (trainVal - testVal);
			result += tmpResult * tmpResult;
		}

		m[trainId] = result; //Coping the result to global memory
		trainId+=blockDim.x;
	}
}


/* Constant memory declaration of the centers memory */
__constant__ float deviceCenters[MAXSIZECENTERS];

__global__ void distanceCenterKernel(float* training, float* test, int nVariables, int trainingSize, int testSize, float* matrix, int chunkTrain, int nChunksTrain, int nNumeric, int selections, int nOutput){
	int trainId = threadIdx.x;
	int testId = blockIdx.x;

	float trainVal, testVal;
	float tmpResult;

	/* Setting pointers to their  positions in memory*/
	float* m = &matrix[testId*(TRAININGCHUNK+selections)];

	//Copy the test values to shared memory
	float* shared = SharedMemory<float>();
	for(int i = threadIdx.x; i < nVariables; i+=blockDim.x)
		shared[i] = test[i*TESTCHUNK+testId];

	__syncthreads();
	for(int r = 0; r < NREP;r++){
		float result = 0;
		int clas = training[(nVariables-1)*TRAININGCHUNK+trainId] * (nOutput - 1);
		float mu = 0;
		float div = 0;
		float testSub;
		float centerSub;

		/* Computing mu value for the instance */
		for(int i = 0; i < nVariables; i++){
			testSub = shared[i] - training[i*TRAININGCHUNK+trainId];
			centerSub = deviceCenters[i*nOutput+clas] - training[i*TRAININGCHUNK+trainId];
			mu += testSub * centerSub;
			div += centerSub * centerSub;
		}
		mu /= div;

		/* Computing distance on the projection */
		for(int i = 0; i < nVariables-1; i++){
			trainVal = training[i*TRAININGCHUNK+trainId] + mu * (deviceCenters[i*nOutput+clas] - training[i*TRAININGCHUNK+trainId]);
			testVal = shared[i];
			tmpResult = (trainVal - testVal);
			result += tmpResult * tmpResult;
		}

		m[trainId] = result;	//Coping the result to global memory
		trainId+=blockDim.x;
	}
}


__global__ void distanceRadiusKernel(float* training, float* test, int nVariables, int trainingSize, int testSize, float* matrix, int chunkTrain, int nChunksTrain, int nNumeric){
	int trainId = threadIdx.x;
	int testId = blockIdx.x;

	int selections = 1;
	float trainVal, testVal;
	float tmpResult;

	/* Setting pointers to their  positions in memory*/
	float* m = &matrix[testId*(TRAININGCHUNK+selections)];

	//Copy the test values to shared memory
	float* shared = SharedMemory<float>();
	for(int i = threadIdx.x; i < nVariables; i+=blockDim.x)
		shared[i] = test[i*TESTCHUNK+testId];

	__syncthreads();
	for(int r = 0; r < NREP;r++){
		float result = 0;
		//Euclidean distance
		for(int i = 0; i < nNumeric; i++){
			trainVal = training[i*TRAININGCHUNK+trainId];
			testVal = shared[i];
			tmpResult = (trainVal - testVal);
			result += tmpResult * tmpResult;
		}
		//A different approach for categorical attributes distance can be used
		//Euclidean distance
		for(int i = nNumeric; i < nVariables-1; i++){
			trainVal = training[i*TRAININGCHUNK+trainId];
			testVal = shared[i];
			//result += (trainVal == testVal) ? 0 : 1;
			tmpResult = (trainVal - testVal);
			result += tmpResult * tmpResult;
		}

		//If the class is the same than the own class the distance is set to a high value
		m[trainId] = shared[nVariables-1] != training[(nVariables-1)*TRAININGCHUNK+trainId] ? result : 9999;
		trainId+=blockDim.x;
	}
}


__global__ void adaptiveDistanceKernel(float* training, float* test, int nVariables, int trainingSize, int testSize, float* matrix, int chunkTrain, int nChunksTrain, int nNumeric, int selections, float* radius){
	int trainId = threadIdx.x;
	int testId = blockIdx.x;

	float trainVal, testVal;
	float tmpResult;

	/* Setting pointers to their  positions in memory*/
	float* m = &matrix[testId*(TRAININGCHUNK+selections)];

	//Copy the test values to shared memory
	float* shared = SharedMemory<float>();
	for(int i = threadIdx.x; i < nVariables; i+=blockDim.x)
		shared[i] = test[i*TESTCHUNK+testId];

	__syncthreads();
	for(int r = 0; r < NREP;r++){
		float result = 0;
		//Euclidean distance
		for(int i = 0; i < nNumeric; i++){
			trainVal = training[i*TRAININGCHUNK+trainId];
			testVal = shared[i];
			tmpResult = (trainVal - testVal);
			result += tmpResult * tmpResult;
		}
		//A different approach for categorical attributes distance can be used
		//Euclidean distance
		for(int i = nNumeric; i < nVariables-1; i++){
			trainVal = training[i*TRAININGCHUNK+trainId];
			testVal = shared[i];
			//result += (trainVal == testVal) ? 0 : 1;
			tmpResult = (trainVal - testVal);
			result += tmpResult * tmpResult;
		}

		// The final distance (square root included) is weighted by the radius of the training instance
		m[trainId] = sqrt(result) / radius[trainId];
		trainId+=blockDim.x;
	}
}

__global__ void distanceOneOutKernel(float* training, float* test, int nVariables, int trainingSize, int testSize, float* matrix, int chunkTrain, int nChunksTrain, int nNumeric, int selections, unsigned long int ofsetTrain, unsigned long int ofsetTest){
	int trainId = threadIdx.x;
	int testId = blockIdx.x;

	float trainVal, testVal;
	float tmpResult;

	/* Setting pointers to their  positions in memory*/
	float* m = &matrix[testId*(TRAININGCHUNK+selections)];

	//Copy the test values to shared memory
	float* shared = SharedMemory<float>();
	for(int i = threadIdx.x; i < nVariables; i+=blockDim.x)
		shared[i] = test[i*TESTCHUNK+testId];

	__syncthreads();
	for(int r = 0; r < NREP;r++){
		float result = 0;
		//Euclidean distance
		for(int i = 0; i < nNumeric; i++){
			trainVal = training[i*TRAININGCHUNK+trainId];
			testVal = shared[i];
			tmpResult = (trainVal - testVal);
			result += tmpResult * tmpResult;
		}
		//A different approach for categorical attributes distance can be used
		//Euclidean distance
		for(int i = nNumeric; i < nVariables-1; i++){
			trainVal = training[i*TRAININGCHUNK+trainId];
			testVal = shared[i];
			//result += (trainVal == testVal) ? 0 : 1;
			tmpResult = (trainVal - testVal);
			result += tmpResult * tmpResult;
		}


		m[trainId] = (trainId+ofsetTrain != testId+ofsetTest) ? result: 9999; // Skipping the instance if has the same index in both sets
		trainId+=blockDim.x;
	}
}

__global__ void distanceSymmetricKernel(float* training, float* test, int nVariables, int trainingSize, int testSize, float* matrix, int chunkTrain, int nChunksTrain, int nNumeric, int selections, int nOutput, float* distances, int* globalVotes, int valid){
	int trainId = threadIdx.x;
	int testId = blockIdx.x;

	float trainVal, testVal;
	float tmpResult;

	/* Setting pointers to their  positions in memory*/
	float* m = &matrix[testId*(TRAININGCHUNK+selections)];

	//Copy the test values to shared memory
	float* shared = SharedMemory<float>();
	for(int i = threadIdx.x; i < nVariables; i+=blockDim.x)
		shared[i] = test[i*TESTCHUNK+testId];

	__syncthreads();
	for(int r = 0; r < NREP;r++){
		float result = 0;
		//Euclidean distance
		for(int i = 0; i < nNumeric; i++){
			trainVal = training[i*TRAININGCHUNK+trainId];
			testVal = shared[i];
			tmpResult = (trainVal - testVal);
			result += tmpResult * tmpResult;
		}
		//A different approach for categorical attributes distance can be used
		//Euclidean distance
		for(int i = nNumeric; i < nVariables-1; i++){
			trainVal = training[i*TRAININGCHUNK+trainId];
			testVal = shared[i];
			//result += (trainVal == testVal) ? 0 : 1;
			tmpResult = (trainVal - testVal);
			result += tmpResult * tmpResult;
		}

		m[trainId] = result; //Coping the result to global memory

		/*Symmetric voting*/
		if((result <= distances[trainId]) && trainId < valid){
			int c = ((int)((training[(nVariables-1)*TRAININGCHUNK+trainId])*(nOutput-1)));
			atomicAdd(&globalVotes[testId*nOutput+c],1); //An atomic operation is used to ensure the result obtained are correct
		}
		trainId+=blockDim.x;
	}

}
