/*
 * CudaSelection.cu
 *
 * Quicksort-based selection kernel implementation
 *
 *      Author: Pablo David Gutiérrez
 *      Author: Miguel Lastra
 *      Author: Jaume Barcardit
 *      Author: José Manuel Benítez
 *      Author: Francisco Herrera
 */

#include "CudaSelection.h"

/** Structure to access shared memory easily
 */
template<class T>
struct SharedMemory
{
    __device__ inline operator       T*()
    {
        extern __shared__ int __smem[];
        return (T*)__smem;
    }

    __device__ inline operator const T*() const
    {
        extern __shared__ int __smem[];
        return (T*)__smem;
    }
};

/* Functions to swap arrays */
__device__ void swap(unsigned long int* &p1, unsigned long int* &p2){
	unsigned long int* tmp = p1;
	p1 = p2;
	p2 = tmp;
}
__device__ void swap(float* &p1, float* &p2){
	float* tmp = p1;
	p1 = p2;
	p2 = tmp;
}

/* Function to compute the median
 * Returns the position of the median
 */
__device__ unsigned long int median(unsigned long int a, unsigned long int b, unsigned long int c, float va, float vb, float vc){
	unsigned long int ret;
	if(va == vb)
		ret = va < vc ? a : c;
	else if(va == vc)
		ret = va < vb ? a : b;
	else if(vb == vc)
		ret = vb < va ? b : a;
	else if((va > vb && va < vc) ||
			(va > vc && va < vb))
		ret = a;
	else if((vb > va && vb < vc) ||
			(vb > vc && vb < va))
		ret = b;
	else if((vc > vb && vc < va) ||
			(vc > va && vc < vb))
		ret = c;
	return ret;
}


__global__ void quickSelection(float* globalRows, int trainingSize, int testSize, int trainChunks, int trainChunk, int testChunk, unsigned long int* returnPossitions, float* returnValues, int selections, float* quickRows, unsigned long int* quickActualIndex, unsigned long int* quickNextIndex, float* quickRowsRight, unsigned long int* quickIndexRight){
	int tid = threadIdx.x;
	int row = blockIdx.x*blockDim.y+threadIdx.y;
	int rowOut = testChunk*TESTCHUNK+row;

	int prevTrain = trainChunk*TRAININGCHUNK;

	if(rowOut >= testSize)
		return;

	/* Setting pointers to their  positions in memory*/
	float* rows = &globalRows[(row)*(TRAININGCHUNK+selections)];
	unsigned long int* index = &quickActualIndex[(row)*(TRAININGCHUNK+selections)];
	float* left = &quickRows[(row)*(TRAININGCHUNK+selections)];
	unsigned long int* leftIndex = &quickNextIndex[(row)*(TRAININGCHUNK+selections)];
	float* right = &quickRowsRight[(row)*(TRAININGCHUNK+selections)];
	unsigned long int* rightIndex = &quickIndexRight[(row)*(TRAININGCHUNK+selections)];


	/* On first iteration the values of the previous neighborhood are not properly set
	 * High values are inserted to avoid erroneous selections
	 */
	if(prevTrain == 0){
		for(int ind = tid; ind < selections; ind += blockDim.x){
			rows[ind+TRAININGCHUNK] = 9999;
		}
	}

	int k = selections;
	int leftPos;
	int rightPos;
	int size = trainingSize - prevTrain < TRAININGCHUNK ? trainingSize - prevTrain: TRAININGCHUNK;

	/* Before start the algorithm is needed to set the values for the indexes to which each distance corresponds*/
	for(int i = tid; i < size; i+=blockDim.x){
		index[i] = prevTrain + i;
	}

	/* Moving current best for last iteration */
	if(trainingSize - prevTrain < TRAININGCHUNK){
		for(int ind = tid; ind < selections; ind += blockDim.x){
			rows[ind + size] = rows[ind + TRAININGCHUNK];
			index[ind + size] = index[ind + TRAININGCHUNK];
		}
	}

	/* Updating size value */
	size += selections;


	float pivot;
	int localPos;
	/* The first pivot possition is set to the farthest neighborhood of the previous iteration
	 * If this is the first iteration the median heuristic is used
	 */
	int pivotPos = prevTrain != 0 ? size-1 : median(0,size-1,size/2,rows[0], rows[size-1],rows[size/2]);

	/* Setting shared memory arrays for each block*/
	float* smValues = SharedMemory<float>();
	unsigned long int* smIndex = (unsigned long int*)&smValues[blockDim.x*blockDim.y];
	smIndex = &smIndex[blockDim.x*threadIdx.y];
	smValues = &smValues[blockDim.x*threadIdx.y];

	unsigned int greater;
	unsigned int lower;
	unsigned int greaterBallot;
	unsigned int lowerBallot;

	// Mask is used to compute the local offset. It is needed to know how many
	// threads with a lane ID less than the actual thread and will write to the same buffer
	unsigned mask;
	asm("mov.u32 %0, %%lanemask_lt;" : "=r"(mask));

	// main quickSelection loop
	while(k > 0){
		leftPos = 0; rightPos = 0;
		pivot = rows[pivotPos];

		//The limit value is computed to ensure all threads enter the for loop
		int limit = size+(blockDim.x-(size%blockDim.x));
		for(int i = tid; i < limit; i+= blockDim.x){
			greater = i<size ? rows[i] > pivot : 0;
			lower =  i<size ? rows[i] <= pivot : 0;

			//__ballot and __popc functions are used to compute the local position within the warp
			lowerBallot = __ballot(lower);
			greaterBallot = __ballot(greater);
			localPos = greater ? __popc(greaterBallot & mask) : __popc(lowerBallot & mask);

			// Local memory non coalescent writings
			if(i < size && lower){
				smValues[localPos] = rows[i];
				smIndex[localPos] = index[i];
			}
			else if(i < size && leftPos < k){
				smValues[__popc(lowerBallot)+localPos] = rows[i];
				smIndex[__popc(lowerBallot)+localPos] = index[i];
			}

			// Global memory coalestcent writings
			if(i < size && tid < __popc(lowerBallot)){
				left[leftPos+tid] = smValues[tid];
				leftIndex[leftPos+tid] = smIndex[tid];
			}
			else if(i < size && leftPos < k){
				right[rightPos+tid-__popc(lowerBallot)] = smValues[tid];
				rightIndex[rightPos+tid-__popc(lowerBallot)] = smIndex[tid];
			}
			leftPos += __popc(lowerBallot);
			rightPos += __popc(greaterBallot);

		}

		//Checking left part size
		if(leftPos > k){
			size = leftPos;

			if(rightPos == 0){ // This happens when all values are equal
				// Selection of the k required values
				for(int i = tid; i < k; i+=blockDim.x){
					returnPossitions[row*selections+(selections-k)+i] = leftIndex[i];
					returnValues[row*selections+(selections-k)+i] = left[i];
				}
				k = 0;
			}
			//Swapping arrays
			swap(rows,left);
			swap(index,leftIndex);
			// Updating pivot
			if(k>0)pivotPos = median(0,size-1,size/2,rows[0], rows[size-1],rows[size/2]);

		}
		else{// Left part smaller than k
			// Saving left part elements (pivot is included)
			for(int i = tid; i < leftPos; i+=blockDim.x){
				returnPossitions[row*selections+(selections-k)+i] = leftIndex[i];
				returnValues[row*selections+(selections-k)+i] = left[i];
			}
			size = rightPos;
			k-=leftPos;	// Updating k value
			//Swapping arrays
			swap(rows,right);
			swap(index,rightIndex);
			// Updating pivot
			if(k>0)pivotPos = median(0,size-1,size/2,rows[0], rows[size-1],rows[size/2]);
		}
	}

	//Writing neighborhood into the distance matrix for next step computation
	for(int ind = tid; ind < selections; ind += blockDim.x){
		globalRows[row * (TRAININGCHUNK+selections) + ind+TRAININGCHUNK] = returnValues[row*selections + ind];
		quickActualIndex[row * (TRAININGCHUNK+selections) + ind+TRAININGCHUNK] = returnPossitions[row*selections + ind];
	}
}


__global__ void quickKomarov(float* globalRows, int trainingSize, int testSize, int trainChunks, int trainChunk, int testChunk, unsigned long int* returnPossitions, float* returnValues, int selections, float* quickRows, unsigned long int* quickActualIndex, unsigned long int* quickNextIndex, float* quickRowsRight, unsigned long int* quickIndexRight){
	int tid = threadIdx.x;
	int row = blockIdx.x;
	int rowOut = testChunk*TESTCHUNK+row;

	int prevTrain = trainChunk*TRAININGCHUNK;

	if(rowOut >= testSize)
		return;

	/* Setting pointers to their  positions in memory*/
	float* rows = &globalRows[(row)*(TRAININGCHUNK+selections)];
	unsigned long int* index = &quickActualIndex[(row)*(TRAININGCHUNK+selections)];
	float* left = &quickRows[(row)*(TRAININGCHUNK+selections)];
	unsigned long int* leftIndex = &quickNextIndex[(row)*(TRAININGCHUNK+selections)];
	float* right = &quickRowsRight[(row)*(TRAININGCHUNK+selections)];
	unsigned long int* rightIndex = &quickIndexRight[(row)*(TRAININGCHUNK+selections)];


	/* On first iteration the values of the previous neighborhood are not properly set
	 * High values are inserted to avoid erroneous selections
	 */
	if(prevTrain == 0){
		for(int ind = tid; ind < selections; ind += blockDim.x){
			rows[ind+TRAININGCHUNK] = 9999;
		}
	}

	int k = selections;
	int leftPos;
	int rightPos;
	int size = trainingSize - prevTrain < TRAININGCHUNK ? trainingSize - prevTrain: TRAININGCHUNK;

	/* Before start the algorithm is needed to set the values for the indexes to which each distance corresponds*/
	for(int i = tid; i < size; i+=blockDim.x){
		index[i] = prevTrain + i;
	}

	/* Moving current best for last iteration */
	if(trainingSize - prevTrain < TRAININGCHUNK){
		for(int ind = tid; ind < selections; ind += blockDim.x){
			rows[ind + size] = rows[ind + TRAININGCHUNK];
			index[ind + size] = index[ind + TRAININGCHUNK];
		}
	}

	__syncthreads();
	/* Updating size value */
	size += selections;

	float pivot;
	int localPos;
	/* The first pivot possition is set to the farthest neighborhood of the previous iteration
	 * If this is the first iteration the median heuristic is used */
	int pivotPos = prevTrain != 0 ? size-1 : median(0,size-1,size/2,rows[0], rows[size-1],rows[size/2]);

	/* Setting shared memory arrays for each block*/
	float* smValues = SharedMemory<float>();
	unsigned long int* smIndex = (unsigned long int*)&smValues[blockDim.x];

	unsigned int greater;
	unsigned int lower;
	unsigned int globGreater;
	unsigned int globLower;
	unsigned int totalLower;
	unsigned int totalGreater;
	unsigned int greaterBallot;
	unsigned int lowerBallot;

	// Mask is used to compute the local offset. I need to know how many
	// threads with a lane ID less than mine are going to write to the same buffer
	unsigned mask;
	asm("mov.u32 %0, %%lanemask_lt;" : "=r"(mask));

	// main quickSelection loop
	while(k > 0){
		leftPos = 0; rightPos = 0;
		pivot = rows[pivotPos];

		//The limit value is computed to ensure all threads enter the for loop
		int limit = size+(blockDim.x-(size%blockDim.x));
		for(int i = tid; i < limit; i+= blockDim.x){
			greater = i<size ? rows[i] > pivot : 0;
			lower =  i<size ? rows[i] <= pivot : 0;

			//__ballot and __popc functions are used to compute the local position within the warp
			lowerBallot = __ballot(lower);
			greaterBallot = __ballot(greater);

			// A prefix sum is used to inform other warps of the result obtained
			if(tid % 32 == 0)
				smValues[tid/32] = __popc(lowerBallot);
			__syncthreads();
			if(tid == 0){
				for(int h = 1; h < blockDim.x /32;h++){
					smValues[h] += smValues[h-1];
				}
			}
			__syncthreads();
			// The writing position in local memory is computed using the results of the prefix sum
			globLower= tid < 32 ? 0 : smValues[(tid/32)-1];
			globGreater = 32*(tid/32) - globLower;
			totalLower = smValues[(blockDim.x/32)-1];
			totalGreater = (i/blockDim.x) < (size/blockDim.x) ? blockDim.x - totalLower : (size%blockDim.x) - totalLower;

			localPos = greater ? globGreater + __popc(greaterBallot & mask) : globLower + __popc(lowerBallot & mask);

			// Local memory non coalescent writings
			__syncthreads();
			if(i < size && lower){
				smValues[localPos] = rows[i];
				smIndex[localPos] = index[i];
			}
			else if(i < size && leftPos < k){
				smValues[totalLower+localPos] = rows[i];
				smIndex[totalLower+localPos] = index[i];
			}

			__syncthreads();
			// Global memory coalestcent writings
			if(i < size && tid < totalLower){
				left[leftPos+tid] = smValues[tid];
				leftIndex[leftPos+tid] = smIndex[tid];
			}
			if(i < size && leftPos < k && tid < totalGreater){
				right[rightPos+tid] = smValues[tid + totalLower];
				rightIndex[rightPos+tid] = smIndex[tid + totalLower];
			}

			leftPos += totalLower;
			rightPos += totalGreater;

		}
		__syncthreads();
		//Checking left part size
		if(leftPos > k){
			size = leftPos;

			if(rightPos == 0){ // This happens when all values are equal
				// Selection of the k required values
				for(int i = tid; i < k; i+=blockDim.x){
					returnPossitions[row*selections+(selections-k)+i] = leftIndex[i];
					returnValues[row*selections+(selections-k)+i] = left[i];
				}
				k = 0;
			}
			//Swapping arrays
			swap(rows,left);
			swap(index,leftIndex);
			// Updating pivot
			if(k>0)pivotPos = median(0,size-1,size/2,rows[0], rows[size-1],rows[size/2]);

		}
		else{// Left part smaller than k
			// Saving left part elements (pivot is included)
			for(int i = tid; i < leftPos; i+=blockDim.x){
				returnPossitions[row*selections+(selections-k)+i] = leftIndex[i];
				returnValues[row*selections+(selections-k)+i] = left[i];
			}
			size = rightPos;
			k-=leftPos; // Updating k value
			//Swapping arrays
			swap(rows,right);
			swap(index,rightIndex);
			// Updating pivot
			if(k>0)
				pivotPos = median(0,size-1,size/2,rows[0], rows[size-1],rows[size/2]);
		}
		__syncthreads();
	}

	//Writing neighborhood into the distance matrix for next step computation
	for(int ind = tid; ind < selections; ind += blockDim.x){
		globalRows[row * (TRAININGCHUNK+selections) + ind+TRAININGCHUNK] = returnValues[row*selections + ind];
		quickActualIndex[row * (TRAININGCHUNK+selections) + ind+TRAININGCHUNK] = returnPossitions[row*selections + ind];
	}
}

