/*
 * DatasetAdaptative.cpp
 *
 * DatasetAdaptative class methods file
 *
 *      Author: Pablo David Gutiérrez
 *      Author: Miguel Lastra
 *      Author: Jaume Barcardit
 *      Author: José Manuel Benítez
 *      Author: Francisco Herrera
 */

#include "DatasetAdaptive.h"

DatasetAdaptive::DatasetAdaptive() : DatasetGPU(){
	deviceRadius = 0;
	radius = 0;
}

DatasetAdaptive::DatasetAdaptive(const DatasetAdaptive& orig){
	nInst = orig.nInst;
	nVariables = orig.nVariables;
	nOutput = orig.nOutput;
	name = orig.name;
	attributes = orig.attributes;
	nNumeric = orig.nNumeric;

	if(orig.variables != 0){
		allocate();
		cudaMemcpy(variables,orig.variables,sizeof(float)*nVariables*nInst,cudaMemcpyHostToHost);
		cudaMemcpy(radius,orig.radius,sizeof(float)*nInst,cudaMemcpyHostToHost);
	}
	else{
		variables = 0;
		variablesGPU = 0;
		e = 0;
		copyStream = 0;
		deviceRadius = 0;
		radius = 0;
	}
}

DatasetAdaptive::~DatasetAdaptive() {
	if(variablesGPU != 0){
		freeGPU();
		cudaFreeHost(variables);
		variables=0;
		cudaFreeHost(radius);
		radius=0;
	}
}


void DatasetAdaptive::computeRadius(){
	DatasetAdaptive aux(*this);
	KNNSkipOwnClass local(this,&aux);
	local.computeDistanceMatrix();
	local.getRadius(radius);
}

void DatasetAdaptive::allocate(){
	DatasetGPU::allocate();
	cudaMallocHost(&radius, sizeof(float) * nInst);
}

void DatasetAdaptive::allocateGPU(){
	DatasetGPU::allocateGPU();
	cudaMalloc(&deviceRadius, sizeof(float) * TRAININGCHUNK);
}

void DatasetAdaptive::freeGPU(){
	DatasetGPU::freeGPU();
	cudaFree(deviceRadius);
}

void DatasetAdaptive::copyChunk(int i, int n){
	int size;
	if(i < n){
		size = (nInst - TRAININGCHUNK*i) < TRAININGCHUNK ? nInst - TRAININGCHUNK*i : TRAININGCHUNK;
		for(int k = 0; k < nVariables; k++)
			cudaMemcpyAsync(variablesGPU+(TRAININGCHUNK*k), variables+(nInst*k + TRAININGCHUNK*i), sizeof(float)*size,cudaMemcpyHostToDevice,copyStream);
		cudaMemcpyAsync(deviceRadius, radius+(TRAININGCHUNK*i), sizeof(float)*size,cudaMemcpyHostToDevice,copyStream);
	}
	cudaEventRecord(e,copyStream);
}

DatasetAdaptive::KNNSkipOwnClass::KNNSkipOwnClass(DatasetGPU* train, DatasetGPU* tst) : KnnGPU(train, tst){

}

DatasetAdaptive::KNNSkipOwnClass::~KNNSkipOwnClass(){

}

void DatasetAdaptive::KNNSkipOwnClass::ComputeDistanceGPU(int trainChunk, int testChunk){
	int nthreads = THREADSMULT;
	int block = TESTCHUNK;
	int shared = training->getNVariables()*sizeof(float);

	distanceRadiusKernel<<<block, nthreads, shared, kernelStream>>>(deviceTraining->getDevicePointer(), deviceTest->getDevicePointer(), training->getNVariables(), trainingSize, testSize, deviceDistanceMatrix, trainChunk, trainingChunks, training->getNNumeric());
}


void DatasetAdaptive::KNNSkipOwnClass::ComputeSelectionGPU(int trainChunk, int testChunk){

	/* Quicksort selection*/
	dim3 blockSize;
	blockSize.x = 32;
	blockSize.y = 4;
	int nblocks = TESTCHUNK/blockSize.y+1*(TESTCHUNK%blockSize.y != 0);
	int totalThreads = blockSize.x*blockSize.y;
	int shared = (sizeof(float)+sizeof(unsigned long int))*totalThreads;
	//in the kernel call 1 is used instead of the k value
	quickSelection<<<nblocks,blockSize,shared,kernelStream>>>(deviceDistanceMatrix, trainingSize, testSize, trainingChunks, trainChunk, testChunk, deviceNeighborIndexFirstStep, deviceNeighborValuesFirstStep, 1, deviceDistanceLeftSelection, deviceNeighborActualIndexSelection, deviceNeighborLeftIndexSelection, deviceDistanceRightSelection, deviceNeighborRightIndexSelection);

}

void DatasetAdaptive::KNNSkipOwnClass::ComputeSqrtGPU(){
	int threads = 128;
	int iterations = 1; //Only one interation is performed as k = 1
	int totalValues = TESTCHUNK*1;
	int blocks = totalValues/(threads*iterations)+(totalValues%(threads*iterations) != 0);

	cuSqrt<<<blocks,threads,0,kernelStream>>>(deviceNeighborValuesFirstStep, deviceNeighborIndexFirstStep, deviceNeighborValues, deviceNeighborIndex, totalValues, iterations);
}


void DatasetAdaptive::KNNSkipOwnClass::copyNeighbors(unsigned long int j){
	cudaMemcpyAsync(neighborIndex+j*TESTCHUNK, deviceNeighborIndex, sizeof(unsigned long int)*TESTCHUNK,cudaMemcpyDeviceToHost);
	cudaMemcpyAsync(neighborValues+j*TESTCHUNK, deviceNeighborValues, sizeof(float)*TESTCHUNK,cudaMemcpyDeviceToHost);
	cudaEventRecord(eCopy);
}

void DatasetAdaptive::KNNSkipOwnClass::getRadius(float* r){
	cudaMemcpy(r,neighborValues,sizeof(float)*training->getNInst(),cudaMemcpyHostToHost);
}
