/*
 * DatasetAdaptative.h
 *
 * DatasetAdaptative class header
 *
 *      Author: Pablo David Gutiérrez
 *      Author: Miguel Lastra
 *      Author: Jaume Barcardit
 *      Author: José Manuel Benítez
 *      Author: Francisco Herrera
 */

#ifndef DATASETADAPTATIVE_H_
#define DATASETADAPTATIVE_H_

#include "DatasetGPU.h"
#include "KnnGPU.h"

/* DatasetAdaptative class inherits from DatasetGPU class adding structures to store the radius of each instance
 * The radius is computed with a modified version of the kNN rule to skip instances from the own class
 */
class DatasetAdaptive: public DatasetGPU {
protected:
	float* radius;			//Radius array
	float* deviceRadius;	//Device radius chunk array

	/* Radius computation kNN rule class*/
	class KNNSkipOwnClass: public KnnGPU {
		/* Copies the neigborhood (radius) to the CPU, in this case the k value is ignored and set to 1 */
		void copyNeighbors(unsigned long int j);
	public:
		/* Constructor */
		KNNSkipOwnClass(DatasetGPU* train, DatasetGPU* tst);
		/* Destructor */
		virtual ~KNNSkipOwnClass();

		/* Wrapper for the modified distance version*/
		virtual void ComputeDistanceGPU(int trainChunk, int testChunk);
		/* Wrapper for the modified parameters selection, in this case the k value is ignored and set to 1 */
		virtual void ComputeSelectionGPU(int trainChunk, int testChunk);
		/* Wrapper for the modified parameters square root, in this case the k value is ignored and set to 1 */
		virtual void ComputeSqrtGPU();

		/* Returns the radius array */
		void getRadius(float* r);
	};
public:
	/* Default constructor */
	DatasetAdaptive();
	/* Copy constructor */
	DatasetAdaptive(const DatasetAdaptive& orig);
	/* Destructor */
	virtual ~DatasetAdaptive();


	/* CPU data allocation function*/
	virtual void allocate();
	/* GPU data allocation function*/
	virtual void allocateGPU();
	/* Free GPU data*/
	virtual void freeGPU();

	/* Copy chunk function for training set, including the radius copy */
	virtual void copyChunk(int i, int n);

	/* Computes the radius */
	void computeRadius();

	/* Returns the GPU radius array */
	inline float* getRadiusPointer() { return deviceRadius; }
	/* Returns the CPU radius array */
	inline float* getRadius() { return radius; }
};

#endif /* DATASETADAPTATIVE_H_ */
