/*
 * DatasetSymmetric.cpp
 *
 * DatasetSymmetric class methods file
 *
 *      Author: Pablo David Gutiérrez
 *      Author: Miguel Lastra
 *      Author: Jaume Barcardit
 *      Author: José Manuel Benítez
 *      Author: Francisco Herrera
 */

#include "DatasetSymmetric.h"

DatasetSymmetric::DatasetSymmetric() :DatasetGPU(){
	distance = 0;
	deviceDistance = 0;
}

DatasetSymmetric::DatasetSymmetric(const DatasetSymmetric& orig){
	nInst = orig.nInst;
	nVariables = orig.nVariables;
	nOutput = orig.nOutput;
	name = orig.name;
	attributes = orig.attributes;
	nNumeric = orig.nNumeric;

	if(orig.variables != 0){
		allocate();
		cudaMemcpy(variables,orig.variables,sizeof(float)*nVariables*nInst,cudaMemcpyHostToHost);
		cudaMemcpy(distance,orig.distance,sizeof(float)*nInst,cudaMemcpyHostToHost);
	}
	else{
		variables = 0;
		variablesGPU = 0;
		e = 0;
		copyStream = 0;
		distance = 0;
		deviceDistance = 0;
	}
}

DatasetSymmetric::~DatasetSymmetric() {
	if(variablesGPU != 0){
		freeGPU();
		cudaFreeHost(variables);
		variables=0;
		cudaFreeHost(distance);
		distance=0;
	}
}


void DatasetSymmetric::computeMaxDistance(){
	DatasetSymmetric aux(*this);
	KNNLeaveOneOut local(this,&aux);
	local.computeDistanceMatrix();
	local.getMaxDistance(distance);
}

void DatasetSymmetric::allocate(){
	DatasetGPU::allocate();
	cudaMallocHost(&distance, sizeof(float) * nInst);
}

void DatasetSymmetric::allocateGPU(){
	DatasetGPU::allocateGPU();
	cudaMalloc(&deviceDistance, sizeof(float) * TRAININGCHUNK);
}

void DatasetSymmetric::freeGPU(){
	DatasetGPU::freeGPU();
	cudaFree(deviceDistance);
}

void DatasetSymmetric::copyChunk(int i, int n){
	if(i < n){
		int size = (nInst - TRAININGCHUNK*i) < TRAININGCHUNK ? nInst - TRAININGCHUNK*i : TRAININGCHUNK;
		for(int k = 0; k < nVariables; k++)
			cudaMemcpyAsync(variablesGPU+(TRAININGCHUNK*k), variables+(nInst*k + TRAININGCHUNK*i), sizeof(float)*size,cudaMemcpyHostToDevice,copyStream);
		cudaMemcpyAsync(deviceDistance, distance+(TRAININGCHUNK*i), sizeof(float)*size,cudaMemcpyHostToDevice,copyStream);
	}
	cudaEventRecord(e,copyStream);
}


DatasetSymmetric::KNNLeaveOneOut::KNNLeaveOneOut(DatasetGPU* train, DatasetGPU* tst) : KnnGPU(train, tst){

}

DatasetSymmetric::KNNLeaveOneOut::~KNNLeaveOneOut(){

}

void DatasetSymmetric::KNNLeaveOneOut::ComputeDistanceGPU(int trainChunk, int testChunk){
	int nthreads = THREADSMULT;
	int block = TESTCHUNK;
	int shared = training->getNVariables()*sizeof(float);

	distanceOneOutKernel<<<block, nthreads, shared, kernelStream>>>(deviceTraining->getDevicePointer(), deviceTest->getDevicePointer(), training->getNVariables(), trainingSize, testSize, deviceDistanceMatrix, trainChunk, trainingChunks, training->getNNumeric(), k, trainChunk*TRAININGCHUNK, testChunk*TESTCHUNK);
}


void DatasetSymmetric::KNNLeaveOneOut::ComputeSqrtGPU(){
	int threads = 128;
	int totalValues = TESTCHUNK;
	int blocks = totalValues/threads+((totalValues%threads) != 0);

	cuGetLast<<<blocks,threads,0,kernelStream>>>(deviceNeighborValuesFirstStep, deviceNeighborIndexFirstStep, deviceNeighborValues, deviceNeighborIndex, totalValues, k);
}

void DatasetSymmetric::KNNLeaveOneOut::copyNeighbors(unsigned long int j){
	cudaMemcpyAsync(neighborIndex+j*TESTCHUNK, deviceNeighborIndex, sizeof(unsigned long int)*TESTCHUNK,cudaMemcpyDeviceToHost);
	cudaMemcpyAsync(neighborValues+j*TESTCHUNK, deviceNeighborValues, sizeof(float)*TESTCHUNK,cudaMemcpyDeviceToHost);
	cudaEventRecord(eCopy);
}


void DatasetSymmetric::KNNLeaveOneOut::getMaxDistance(float* r){
	cudaMemcpy(r,neighborValues,sizeof(float)*training->getNInst(),cudaMemcpyHostToHost);
}
