/*
 * DatasetSymmetric.h
 *
 * DatasetSymmetric class header
 *
 *      Author: Pablo David Gutiérrez
 *      Author: Miguel Lastra
 *      Author: Jaume Barcardit
 *      Author: José Manuel Benítez
 *      Author: Francisco Herrera
 */

#ifndef DATASETSYMMETRIC_H_
#define DATASETSYMMETRIC_H_

#include "DatasetGPU.h"
#include "KnnGPU.h"
#include "const.h"

/* DatasetSymmetric class inherits from DatasetGPU class adding structures to store the distance to the farthest element of each instance
 * The distance to the farthest element is computed with a modified version of the kNN rule that computes the neighborhood skipping the computation
 * if training and test instances have the same index
 */
class DatasetSymmetric: public DatasetGPU {
protected:
	float* distance;		//Distance Array
	float* deviceDistance;	//Device distance chunk array

	/* Farthest distance computation kNN rule class*/
	class KNNLeaveOneOut : public KnnGPU {
	private:
		/* Modified copy, only the farthest neighbors are copied*/
		virtual void copyNeighbors(unsigned long int j);
	public:
		/* Constructor */
		KNNLeaveOneOut(DatasetGPU* train, DatasetGPU* tst);
		/* Destructor */
		virtual ~KNNLeaveOneOut();

		/* Wrapper for the modified distance version*/
		virtual void ComputeDistanceGPU(int trainChunk, int testChunk);
		/* Wrapper for the modified square root, only the farthest neighbors are copied
		 * the square root is not computed because the comparison for the symmetric vote is done during the distance matrix computation, also without SQRT */
		virtual void ComputeSqrtGPU();

		/* Returns the distance array */
		void getMaxDistance(float* r);
	};
public:
	/* Default constructor */
	DatasetSymmetric();
	/* Copy constructor */
	DatasetSymmetric(const DatasetSymmetric& orig);
	/* Destructor */
	virtual ~DatasetSymmetric();

	/* CPU data allocation function*/
	virtual void allocate();
	/* GPU data allocation function*/
	virtual void allocateGPU();
	/* Free GPU data*/
	virtual void freeGPU();

	/* Copy chunk function for training set, including the distance copy */
	virtual void copyChunk(int i, int n);

	/* Computes the fartherst distance*/
	void computeMaxDistance();

	/* Returns the GPU distance array */
	inline float* getDistanceDevicePointer() { return deviceDistance; }
	/* Returns the CPU distance array */
	inline float* getDistancePointer() { return distance; }
	/* Returns de farthest distance for a specific instance */
	inline float getDistance(int inst) { return distance[inst]; }
};

#endif /* DATASETSYMMETRIC_H_ */
