/*
 * KNNSymmetricGPU.cpp
 *
 * KNNSymmetricGPU class methods file
 *
 *      Author: Pablo David Gutiérrez
 *      Author: Miguel Lastra
 *      Author: Jaume Barcardit
 *      Author: José Manuel Benítez
 *      Author: Francisco Herrera
 */

#include "KNNSymmetricGPU.h"

KNNSymmetricGPU::KNNSymmetricGPU(DatasetGPU* train, DatasetGPU* tst) : KnnGPU(train,tst) {
	globalVotes = 0;
	deviceGlobalVotes = 0;
	deviceGlobalVotesCopy = 0;
}

KNNSymmetricGPU::~KNNSymmetricGPU() {
	if(globalVotes != 0){
		freeGlobalVotes();
		globalVotes = 0;
		deviceGlobalVotes = 0;
		deviceGlobalVotesCopy = 0;
	}
}


void KNNSymmetricGPU::allocateGlobalVotes(){
	cudaMallocHost(&globalVotes, sizeof(int) * test->getNInst() * test->getNOutput());
	cudaMalloc(&deviceGlobalVotes, sizeof(int) * TESTCHUNK * test->getNOutput());
	cudaMalloc(&deviceGlobalVotesCopy, sizeof(int) * TESTCHUNK * test->getNOutput());
}

void KNNSymmetricGPU::freeGlobalVotes(){
	cudaFree(deviceGlobalVotes);
	cudaFree(deviceGlobalVotesCopy);
	cudaFreeHost(globalVotes);
}

void KNNSymmetricGPU::initialize(){
	allocateGlobalVotes();
	((DatasetSymmetric*)training)->computeMaxDistance();
}

void KNNSymmetricGPU::copyNeighbors(unsigned long int j){
	cudaMemcpyAsync(globalVotes+j*test->getNOutput()*TESTCHUNK, deviceGlobalVotesCopy, sizeof(int)* test->getNOutput() *TESTCHUNK,cudaMemcpyDeviceToHost);
	KnnGPU::copyNeighbors(j);
}

void KNNSymmetricGPU::ComputeDistanceGPU(int trainChunk, int testChunk){
	//number of valid instances of the chunk
	int valid = (training->getNInst() - TRAININGCHUNK*trainChunk) < TRAININGCHUNK ? training->getNInst() - TRAININGCHUNK*trainChunk : TRAININGCHUNK;

	/* For the first training chunk the voting structre is reset to 0*/
	if(trainChunk == 0){
		cudaMemset(deviceGlobalVotes,0,sizeof(int)*TESTCHUNK*training->getNOutput());
	}

	int nthreads = THREADSMULT;
	int block = TESTCHUNK;
	int shared = training->getNVariables()*sizeof(float);

	distanceSymmetricKernel<<<block, nthreads, shared, kernelStream>>>(deviceTraining->getDevicePointer(), deviceTest->getDevicePointer(), training->getNVariables(), trainingSize, testSize, deviceDistanceMatrix, trainChunk, trainingChunks, training->getNNumeric(), k, test->getNOutput(), ((DatasetSymmetric*)training)->getDistanceDevicePointer(), deviceGlobalVotes,valid);

}

void KNNSymmetricGPU::ComputeSqrtGPU(){
	//copying the voting structure to its final destination
	cudaMemcpy(deviceGlobalVotesCopy,deviceGlobalVotes,sizeof(int)*TESTCHUNK*test->getNOutput(),cudaMemcpyDeviceToDevice);

	int threads = 128;
	int iterations = k;
	int totalValues = TESTCHUNK*k;
	int blocks = totalValues/(threads*iterations)+(totalValues%(threads*iterations) != 0);

	cuSqrt<<<blocks,threads,0,kernelStream>>>(deviceNeighborValuesFirstStep, deviceNeighborIndexFirstStep, deviceNeighborValues, deviceNeighborIndex, totalValues, iterations);
}

float KNNSymmetricGPU::computeAcuracy(){

	float acc = 0;
	DatasetSymmetric* tr = (DatasetSymmetric*)training;

	for(unsigned long int h = 0; h < testSize; h++){
		vector<int> votes;
		votes.resize(training->getNOutput(),0);
		for(int i = 0; i < training->getNOutput(); i++){
			votes[i] = globalVotes[h*training->getNOutput()+i];
		}
		for(int i = 0; i < k; i++){
			//if the instance already voted we skip it
			if(tr->getDistance(neighborIndex[h*k+i]) >= neighborValues[h*k+i])
				votes[training->getClassNumber(neighborIndex[h*k+i])]++;

		}
		int max = -1;
		int clas = -1;
		for(int i = 0; i < training->getNOutput(); i++){
			if(votes[i] > max){
				clas = i;
				max = votes[i];
			}
		}

		if(test->getClass(h) == test->getOutputValue(clas))
			acc++;
	}

	return acc/testSize;
}
