/*
 * Knn.cpp
 *
 * Knn class methods file
 *
 *      Author: Pablo David Gutiérrez
 *      Author: Miguel Lastra
 *      Author: Jaume Barcardit
 *      Author: José Manuel Benítez
 *      Author: Francisco Herrera
 */

#include "Knn.h"
#include <cmath>
#include <iostream>

int Knn::k;

Knn::Knn() {
	training = 0;
	test = 0;
	trainingSize = 0;
	testSize = 0;
	distanceMatrix = 0;
}

Knn::Knn(Dataset* train, Dataset* tst) {
	training = train;
	test = tst;
	distanceMatrix = new float[(unsigned long int)training->getNInst() * test->getNInst()];
	trainingSize = training->getNInst();
	testSize = test->getNInst();

}

Knn::~Knn() {
	if(distanceMatrix != 0)
		delete[] distanceMatrix;
}

float Knn::distance(unsigned long int trainIndex, unsigned long int testIndex){
	float d = 0;
	//Euclidean distance computation
	for(int i = 0; i < training->getNNumeric(); i++){
		d += (training->getValue(trainIndex,i) - test->getValue(testIndex,i)) * (training->getValue(trainIndex,i) - test->getValue(testIndex,i));
	}
	for(int i = training->getNNumeric(); i < training->getNVariables()-1; i++){
		// A different distance measure can be set for categorical attributes
		d += (training->getValue(trainIndex,i) - test->getValue(testIndex,i)) * (training->getValue(trainIndex,i) - test->getValue(testIndex,i));
		//d += training->getValue(trainIndex,i)==test->getValue(testIndex,i) ? 0 : 1;
	}

	return sqrt(d);
}

unsigned long int Knn::median(unsigned long int a, unsigned long int b, unsigned long int c){
	unsigned long int ret;
	if((distanceMatrix[a] > distanceMatrix[b] && distanceMatrix[a] <= distanceMatrix[c]) ||
			(distanceMatrix[a] > distanceMatrix[c] && distanceMatrix[a] <= distanceMatrix[b]))
		ret = a;
	else if((distanceMatrix[b] > distanceMatrix[a] && distanceMatrix[b] <= distanceMatrix[c]) ||
			(distanceMatrix[b] > distanceMatrix[c] && distanceMatrix[b] <= distanceMatrix[a]))
		ret = b;
	else if((distanceMatrix[c] > distanceMatrix[b] && distanceMatrix[c] <= distanceMatrix[a]) ||
			(distanceMatrix[c] > distanceMatrix[a] && distanceMatrix[c] <= distanceMatrix[b]))
		ret = c;
	else if(distanceMatrix[a] == distanceMatrix[b] || distanceMatrix[a] == distanceMatrix[c])
		ret = a;
	else if(distanceMatrix[b] == distanceMatrix[c])
		ret = c;
	return ret;
}

vector<unsigned long int> Knn::selection(unsigned long int inst){
	unsigned long int left = inst*trainingSize;
	unsigned long int right = left+trainingSize;
	vector<unsigned long int> ret;

	if(k < 5){ //Insertion selection
		ret.push_back(left);
		vector<unsigned long int>::iterator it ;
		for(unsigned long int i = left+1; i < left+k; i++){
			for(it = ret.begin(); it != ret.end(); it++){
				if(distanceMatrix[i] < distanceMatrix[*it]){
					ret.insert(it,i);
					break;
				}
			}
			if(it==ret.end())
				ret.push_back(i);
		}

		for(unsigned long int i = left + k; i < right; i++){
			for(it = ret.begin(); it != ret.begin()+k; it++){
				if(distanceMatrix[i] < distanceMatrix[*it]){
					ret.insert(it,i);
					break;
				}
			}
		}
		ret.resize(k);
	}
	else{ //quickSort selection
		unsigned long int pivot = median(left, right-1, left + (right - left)/2);
		vector<unsigned long int> lower;
		vector<unsigned long int> higher;

		for(unsigned long int i = left; i < right; i++)
			if(i != pivot){
				if(distanceMatrix[i] <= distanceMatrix[pivot])
					lower.push_back(i);
				else
					higher.push_back(i);
			}

		if(lower.size() > k){
			ret = quick(lower,0);
		}
		else{
			if(lower.size() == k-1){
				lower.push_back(pivot);
			}
			else if(lower.size() < k-1){
				lower.push_back(pivot);
				vector<unsigned long int> tmp = quick(higher, lower.size());
				lower.insert(lower.end(),tmp.begin(),tmp.end());
			}
			ret = lower;
		}
	}

	for(int i = 0; i < k; i++){
		ret[i]=ret[i]%trainingSize;
	}
	return ret;
}


vector<unsigned long int> Knn::quick(vector<unsigned long int> index, int acum){
	unsigned long int left = 0;
	unsigned long int right = index.size();
	unsigned long int pivot = median(left, right-1, left + (right - left)/2);
	vector<unsigned long int> lower;
	vector<unsigned long int> higher;

	for(int i = 0; i < right; i++)
		if(i != pivot){
			if(distanceMatrix[index[i]] <= distanceMatrix[index[pivot]])
				lower.push_back(index[i]);
			else
				higher.push_back(index[i]);
		}

	if((lower.size()+acum) > k){
		return quick(lower,acum);
	}
	else{
		if((lower.size()+acum) == k-1){
			lower.push_back(index[pivot]);
		}
		else if(lower.size()+acum < k-1){
			lower.push_back(index[pivot]);
			vector<unsigned long int> tmp = quick(higher,acum+lower.size());
			lower.insert(lower.end(),tmp.begin(),tmp.end());
		}
		return lower;
	}
}

void Knn::computeDistanceMatrix(){
	for(unsigned long int i = 0; i < testSize; i++){
		for(unsigned long int j = 0; j < trainingSize; j++){
			distanceMatrix[i*trainingSize+j] = distance(j, i);
		}
	}
}

float Knn::computeAcuracy(){
	float acc = 0;

	for(unsigned long int h = 0; h < testSize; h++){
		vector<unsigned long int> index = selection(h);
		vector<int> votes;
		votes.resize(training->getNOutput(),0);
		for(int i = 0; i < k; i++){
			votes[training->getClassNumber(index[i])]++;
		}
		int max = -1;
		int clas = -1;
		for(int i = 0; i < training->getNOutput(); i++){
			if(votes[i] > max){
				clas = i;
				max = votes[i];
			}
		}

		if(test->getClass(h) == test->getOutputValue(clas))
			acc++;
	}

	return acc/testSize;
}

void Knn::initialize(){}
