/*
 * KnnGPU.cpp
 *
 * KnnGPU class methods file
 *
 *      Author: Pablo David Gutiérrez
 *      Author: Miguel Lastra
 *      Author: Jaume Barcardit
 *      Author: José Manuel Benítez
 *      Author: Francisco Herrera
 */

#include "KnnGPU.h"

float* KnnGPU::deviceDistanceMatrix;
float* KnnGPU::deviceNeighborValues;
float* KnnGPU::deviceNeighborValuesFirstStep;
unsigned long int* KnnGPU::deviceNeighborIndex;
unsigned long int* KnnGPU::deviceNeighborIndexFirstStep;
unsigned long int* KnnGPU::deviceNeighborActualIndexSelection;
unsigned long int* KnnGPU::deviceNeighborLeftIndexSelection;
unsigned long int* KnnGPU::deviceNeighborRightIndexSelection;
float* KnnGPU::deviceDistanceLeftSelection;
float* KnnGPU::deviceDistanceRightSelection;

KnnGPU::KnnGPU(DatasetGPU* train, DatasetGPU* tst) {
	// TODO Auto-generated constructor stub

	training = (Dataset*)train;
	test = (Dataset*)tst;
	trainingSize = training->getNInst();
	testSize = test->getNInst();

	deviceTraining = train;
	deviceTest = tst;

	trainingChunks = trainingSize / TRAININGCHUNK + (trainingSize % TRAININGCHUNK != 0);
	testChunks = testSize / TESTCHUNK + (testSize % TESTCHUNK != 0);

	//cudaMallocHost(&distanceMatrix, sizeof(float)*(trainingChunks*TRAININGCHUNK) * (testChunks*TESTCHUNK));
	distanceMatrix = 0;

	cudaMallocHost(&neighborValues,sizeof(float)*MAXK*testChunks*TESTCHUNK);
	cudaMallocHost(&neighborIndex,sizeof(unsigned long int)*MAXK*testChunks*TESTCHUNK);

	cudaEventCreate(&e);
	cudaEventCreate(&eCopy);
	cudaStreamCreate(&copyStream);
	cudaStreamCreate(&kernelStream);

}

KnnGPU::~KnnGPU() {
	// TODO Auto-generated destructor stub

	if(distanceMatrix != 0)
		cudaFreeHost(distanceMatrix);
	distanceMatrix = 0;

	cudaFreeHost(neighborValues);
	cudaFreeHost(neighborIndex);

	cudaEventDestroy(e);
	cudaEventDestroy(eCopy);
	cudaStreamDestroy(copyStream);
	cudaStreamDestroy(kernelStream);
}

void KnnGPU::allocateGPU(){
	cudaMalloc(&deviceDistanceMatrix,sizeof(float)*(TRAININGCHUNK+k)*TESTCHUNK);
	cudaMalloc(&deviceNeighborValues,sizeof(float)*k*TESTCHUNK);
	cudaMalloc(&deviceNeighborIndex,sizeof(unsigned long int)*k*TESTCHUNK);
	cudaMalloc(&deviceNeighborValuesFirstStep,sizeof(float)*k*TESTCHUNK);
	cudaMalloc(&deviceNeighborIndexFirstStep,sizeof(unsigned long int)*k*TESTCHUNK);

	/*quick selection*/
	cudaMalloc(&deviceDistanceLeftSelection,sizeof(float)*(TRAININGCHUNK+k)*TESTCHUNK);
	cudaMalloc(&deviceDistanceRightSelection,sizeof(float)*(TRAININGCHUNK+k)*TESTCHUNK);
	cudaMalloc(&deviceNeighborActualIndexSelection,sizeof(unsigned long int)*(TRAININGCHUNK+k)*TESTCHUNK);
	cudaMalloc(&deviceNeighborLeftIndexSelection,sizeof(unsigned long int)*(TRAININGCHUNK+k)*TESTCHUNK);
	cudaMalloc(&deviceNeighborRightIndexSelection,sizeof(unsigned long int)*(TRAININGCHUNK+k)*TESTCHUNK);

}
void KnnGPU::freeGPU(){
	if(deviceDistanceMatrix){
		cudaFree(deviceDistanceMatrix);
		cudaFree(deviceNeighborIndex);
		cudaFree(deviceNeighborValues);
		cudaFree(deviceNeighborIndexFirstStep);
		cudaFree(deviceNeighborValuesFirstStep);

		/* quick selection */
		cudaFree(deviceDistanceLeftSelection);
		cudaFree(deviceDistanceRightSelection);
		cudaFree(deviceNeighborActualIndexSelection);
		cudaFree(deviceNeighborLeftIndexSelection);
		cudaFree(deviceNeighborRightIndexSelection);


		deviceNeighborValues = 0;
		deviceNeighborIndex = 0;
		deviceNeighborValuesFirstStep = 0;
		deviceNeighborIndexFirstStep = 0;
	}

}

void KnnGPU::setK(int i){
	k = i;
	freeGPU();
	allocateGPU();

}

void KnnGPU::copyNeighbors(unsigned long int j){
	cudaMemcpyAsync(neighborIndex+j*k*TESTCHUNK, deviceNeighborIndex, sizeof(unsigned long int)*k*TESTCHUNK,cudaMemcpyDeviceToHost);
	cudaMemcpyAsync(neighborValues+j*k*TESTCHUNK, deviceNeighborValues, sizeof(float)*k*TESTCHUNK,cudaMemcpyDeviceToHost);
	cudaEventRecord(eCopy);
}

void KnnGPU::computeDistanceMatrix(){
	//Copy the first train and test chunks
	deviceTest->copyChunkTest(0,testChunks);
	deviceTraining->copyChunk(0,trainingChunks);

	for(int j = 0; j < testChunks; j++){
		deviceTest->sync();	//Wait for test chunk copy
		for(int i = 0; i < trainingChunks; i++){
			deviceTraining->sync();	//Wait for traning chunk copy
			//Distance kernel call
			ComputeDistanceGPU(i, j);
			cudaEventRecord(e);
			cudaEventSynchronize(e);
			//wait the kernel to finish and copy the next chunks
			if(i == trainingChunks-1){
				deviceTest->copyChunkTest(j+1 ,testChunks);
				deviceTraining->copyChunk(0,trainingChunks);
			}
			else{
				deviceTraining->copyChunk(i+1 ,trainingChunks);
			}
			//Selection kernel call
			ComputeSelectionGPU(i, j);

			cudaEventRecord(e);
			cudaEventSynchronize(e);	//wait for selection kernel to finish
		}
		cudaEventSynchronize(eCopy);	//Ensure previous copy has finished
		//SQRT kernel call
		ComputeSqrtGPU();
		copyNeighbors(j);	//Copy the neighborhood to CPU
	}
}

float KnnGPU::computeAcuracy(){
	float acc = 0;
	cudaEventSynchronize(eCopy);	//Ensure last copy has finished

	for(unsigned long int h = 0; h < testSize; h++){
		vector<int> votes;
		votes.resize(training->getNOutput(),0);
		for(int i = 0; i < k; i++){
			votes[training->getClassNumber(neighborIndex[h*k+i])]++;
		}
		int max = -1;
		int clas = -1;
		for(int i = 0; i < training->getNOutput(); i++){
			if(votes[i] > max){
				clas = i;
				max = votes[i];
			}
		}

		if(test->getClass(h) == test->getOutputValue(clas))
			acc++;
	}

	return acc/testSize;
}

void KnnGPU::ComputeDistanceGPU(int trainChunk, int testChunk){
	int nthreads = THREADSMULT;
	int block = TESTCHUNK;
	int shared = training->getNVariables()*sizeof(float);

	distanceKernel<<<block, nthreads, shared, kernelStream>>>(deviceTraining->getDevicePointer(), deviceTest->getDevicePointer(), training->getNVariables(), trainingSize, testSize, deviceDistanceMatrix, trainChunk, trainingChunks, training->getNNumeric(), k);
}


void KnnGPU::ComputeSelectionGPU(int trainChunk, int testChunk){
	/* Quicksort selection*/
	dim3 blockSize;
	blockSize.x = 32;
	blockSize.y = 4;
	int nblocks = TESTCHUNK/blockSize.y+1*(TESTCHUNK%blockSize.y != 0);
	int totalThreads = blockSize.x*blockSize.y;
	int shared = (sizeof(float)+sizeof(unsigned long int))*totalThreads;
	quickSelection<<<nblocks,blockSize,shared,kernelStream>>>(deviceDistanceMatrix, trainingSize, testSize, trainingChunks, trainChunk, testChunk, deviceNeighborIndexFirstStep, deviceNeighborValuesFirstStep, k, deviceDistanceLeftSelection, deviceNeighborActualIndexSelection, deviceNeighborLeftIndexSelection, deviceDistanceRightSelection, deviceNeighborRightIndexSelection);

}

void KnnGPU::ComputeSqrtGPU(){
	int threads = 128;
	int iterations = k;
	int totalValues = TESTCHUNK*k;
	int blocks = totalValues/(threads*iterations)+(totalValues%(threads*iterations) != 0);

	cuSqrt<<<blocks,threads,0,kernelStream>>>(deviceNeighborValuesFirstStep, deviceNeighborIndexFirstStep, deviceNeighborValues, deviceNeighborIndex, totalValues, iterations);
}

void KnnGPU::initialize(){}

