/*
 * KnnGPU.h
 *
 * KnnGPU class header
 *
 *      Author: Pablo David Gutiérrez
 *      Author: Miguel Lastra
 *      Author: Jaume Barcardit
 *      Author: José Manuel Benítez
 *      Author: Francisco Herrera
 */

#ifndef KNNGPU_H_
#define KNNGPU_H_

#include "Knn.h"
#include "DatasetGPU.h"
#include "CudaDistanceMatrix.h"
#include "CudaSelection.h"
#include "CudaSqrt.h"
#include "const.h"
#include "util.h"

/* KnnGPU class inherits from Knn class adding GPU structures and allocations */
class KnnGPU : public Knn{
protected:
	DatasetGPU* deviceTraining;				//Training set with GPU structures
	DatasetGPU* deviceTest;					//Test set with GPU structures
	static float* deviceDistanceMatrix;		//Distance matrix structure on GPU

	float* neighborValues;							//Neigborhood distance values
	static float* deviceNeighborValues;				//Final neigborhood distance values on GPU, used for copies
	static float* deviceNeighborValuesFirstStep;	//Neigborhood distance values on GPU
	static float* deviceDistanceLeftSelection;		//Left part of the distance array on GPU
	static float* deviceDistanceRightSelection;		//Right part of the distance array on GPU

	unsigned long int* neighborIndex;								//Neighborhood indexes
	static unsigned long int* deviceNeighborIndex;					//Final neighborhood indexes on GPU, used for copies
	static unsigned long int* deviceNeighborIndexFirstStep;			//Neighborhood indexes on GPU
	static unsigned long int* deviceNeighborActualIndexSelection;	//Array indexes on GPU
	static unsigned long int* deviceNeighborLeftIndexSelection;		//Left part of the array indexes on GPU
	static unsigned long int* deviceNeighborRightIndexSelection;	//Right part of the array indexes on GPU


	int trainingChunks;	//Number of training chunks
	int testChunks;		//Number of test chunks

	cudaEvent_t e;				//Event for kernel call synchronization
	cudaEvent_t eCopy;			//Event for memory copies synchronization
	cudaStream_t copyStream;	//Stream for copies
	cudaStream_t kernelStream;	//Stream for kernels

	/* Allocates GPU memory*/
	static void allocateGPU();
	/* Frees GPU memory*/
	static void freeGPU();

	/* Copies the neighborhood from GPU memory to CPU memory*/
	virtual void copyNeighbors(unsigned long int j);

public:
	/* Creates a kNN to the referred training and test sets*/
	KnnGPU(DatasetGPU* train, DatasetGPU* tst);
	/* Destructor */
	virtual ~KnnGPU();

	/* Sets the value of k
	 * The GPU structured are released and reallocated when k changes
	 * */
	static void setK(int i);

	/* Computes the distance matrix and performs the selection on GPU
	 * Keeps the name of the super class function
	 * */
	virtual void computeDistanceMatrix();

	/* Classifies the test set with the selected neighborhood
	 * Returns the accuracy value */
	virtual float computeAcuracy();

	/* Distance kernel wrapper function */
	virtual void ComputeDistanceGPU(int trainChunk, int testChunk);
	/* Selection kernel wrapper function */
	virtual void ComputeSelectionGPU(int trainChunk, int testChunk);
	/* SQRT kernel wrapper function */
	virtual void ComputeSqrtGPU();

	/* If required, computes information before the kNN computation*/
	virtual void initialize();
};

#endif /* KNNGPU_H_ */
