/**
 * main.cpp
 *
 * Main program file
 *
 *      Author: Pablo David Gutiérrez
 *      Author: Miguel Lastra
 *      Author: Jaume Barcardit
 *      Author: José Manuel Benítez
 *      Author: Francisco Herrera
 */
#include "Knn.h"
#include "Dataset.h"
#include "KnnGPU.h"
#include "KnnKomarov.h"
#include "DatasetGPU.h"
#include "DatasetCenter.h"
#include "CenterKnnGPU.h"
#include "DatasetAdaptive.h"
#include "KNNAdaptiveGPU.h"
#include "DatasetSymmetric.h"
#include "KNNSymmetricGPU.h"
#include "util.h"
#include <cuda_runtime.h>
#include <iostream>

/**
 * Main host function.
 */
int main(int argc, char** argv) {

	if(argc != 5){
		cerr<<"Usage: KNN algorithm k trainingDataset testDataset"<<endl;
		cerr<<"algorithm options:"<<endl;
		cerr<<"0: CPU kNN rule"<<endl;
		cerr<<"1: GPU kNN rule"<<endl;
		cerr<<"2: GPU Center kNN rule"<<endl;
		cerr<<"3: GPU kNN Adaptive rule"<<endl;
		cerr<<"4: GPU Symmetric kNN rule"<<endl;
		cerr<<"5: GPU kNN rule using Komarov's proposal"<<endl;
		exit(-1);
	}


	//Time meausring variables
	timespec start, finish, dif;
	float time;

	float accuracy;
	int type = atoi(argv[1]);
	int k = atoi(argv[2]);

	clock_gettime(CLOCK_REALTIME,&start);
	Dataset *training, *test;
	Knn *knn;

	// Creating datasets for the specified type
	switch(type){
		case 5: //GPU kNN Komarov's
		case 1: //GPU kNN
			training = new DatasetGPU();
			test = new DatasetGPU();
			break;
		case 2: //GPU Center kNN
			training = new DatasetCenter();
			test = new DatasetCenter();
			break;
		case 3: //GPU kNN Adaptive
			training = new DatasetAdaptive();
			test = new DatasetAdaptive();
			break;
		case 4: //GPU kNN Symmetric
			training = new DatasetSymmetric();
			test = new DatasetSymmetric();
			break;
		default: //CPU kNN
			training = new Dataset();
			test = new Dataset();
			break;
	}

	// Loading datasets
	training->load(argv[3]);
	test->load(argv[4]);

	// Creating kNN algorithm of the specified type
	switch(type){
		case 1: //GPU kNN
			KnnGPU::setK(k);
			knn = new KnnGPU((DatasetGPU*)training,(DatasetGPU*)test);
			break;
		case 2: //GPU Center kNN
			KnnGPU::setK(k);
			knn = new CenterKnnGPU((DatasetGPU*)training,(DatasetGPU*)test);
			break;
		case 3: //GPU kNN Adaptive
			KnnGPU::setK(k);
			knn = new KNNAdaptiveGPU((DatasetGPU*)training,(DatasetGPU*)test);
			break;
		case 4: //GPU kNN Symmetric
			KnnGPU::setK(k);
			knn = new KNNSymmetricGPU((DatasetGPU*)training,(DatasetGPU*)test);
			break;
		case 5: //GPU kNN
			KnnGPU::setK(k);
			knn = new KnnKomarov((DatasetGPU*)training,(DatasetGPU*)test);
			break;
		default: //CPU kNN
			Knn::setK(k);
			knn = new Knn(training,test);
			break;
	}

	cout<<"Training set file: "<<argv[3]<<endl;
	cout<<"Test set file: "<<argv[4]<<endl;
	cout<<"Number of variables: "<<training->getNVariables()-1<<endl; //-1 because the output is also included
	cout<<"Number of Training instances: "<<training->getNInst()<<endl;
	cout<<"Number of Test instances: "<<test->getNInst()<<endl;
	cout<<"k value : "<<k<<endl;

	//Initializing algorithm
	knn->initialize();

	clock_gettime(CLOCK_REALTIME,&finish);
	dif = diff(start, finish);
	time = dif.tv_sec+dif.tv_nsec/1000000000.0; //s
	cout<<"Initialization time "<<time<<" s"<<endl;

	//Computing distance matrix and neighborhood (neighborhood only in GPU version)
	clock_gettime(CLOCK_REALTIME,&start);
	knn->computeDistanceMatrix();
	clock_gettime(CLOCK_REALTIME,&finish);
	dif = diff(start, finish);
	time = dif.tv_sec+dif.tv_nsec/1000000000.0; //s
	if(type == 0)
		cout<<"Distance computation time: "<<time<<" s"<<endl;
	else
		cout<<"Distance and neighborhood computation time: "<<time<<" s"<<endl;

	//Computing neighborhood (CPU version only) and Classifying the instances
	clock_gettime(CLOCK_REALTIME,&start);
	accuracy = knn->computeAcuracy();
	clock_gettime(CLOCK_REALTIME,&finish);
	dif = diff(start, finish);
	time = dif.tv_sec+dif.tv_nsec/1000000000.0; //s
	if(type == 0)
		cout<<"Selection computation and classification time: "<<time<<" s"<<endl;
	else
		cout<<"Classification time (performed on CPU): "<<time<<" s"<<endl;

	cout<<"Accuracy: "<<accuracy<<endl;

	//cleaning up
	delete training;
	delete test;
	delete knn;
	cudaDeviceReset();

	return 0;
}
